import copy
from typing import TYPE_CHECKING, List, Optional

from cleo.commands.command import Command
from cleo.helpers import option
from cleo.io.io import IO

from poetry_workspace.errors import WorkspaceError
from poetry_workspace.formatter import WorkspaceFormatter
from poetry_workspace.graph import DependencyGraph
from poetry_workspace.workspace import Workspace

if TYPE_CHECKING:
    from poetry.core.packages.package import Package
    from poetry.poetry import Poetry


class WorkspaceCommand(Command):
    options = [
        option("project", "p", "Select project name to include.", flag=False, multiple=True),
        option("include-dependencies", "d", "Include transitive dependencies of the selected projects."),
        option(
            "include-reverse-dependencies", "r", "Include projects that transitively depend on the selected projects."
        ),
        option("since", None, "Select projects that have changed since this reference.", flag=False),
    ]

    _workspace: Optional["Workspace"]
    _graph: Optional[DependencyGraph]

    def __init__(self) -> None:
        self._workspace = None
        self._graph = None

        super().__init__()

    @property
    def workspace(self) -> "Workspace":
        if not self._workspace:
            raise WorkspaceError("workspace not set")
        return self._workspace

    def set_workspace(self, workspace: "Workspace") -> None:
        self._workspace = workspace
        self._graph = None

    def handle(self) -> int:
        if not self.workspace:
            self.line("The 'workspace' command is only supported from within a workspace", style="error")
            return 1

        exit_code = self.pre_handle()
        if exit_code:
            return exit_code

        for project in self.selected_projects():
            poetry = self.workspace.find_project(project.name)
            if not poetry:
                self.line(f"Project {project.name} not found in workspace", style="error")
                return 1

            exit_code = self.handle_each(poetry, self._io_for_project(project.name))
            if exit_code:
                return exit_code

        exit_code = self.post_handle()
        if exit_code:
            return exit_code

        return 0

    def pre_handle(self) -> int:
        return 0

    def handle_each(self, poetry: "Poetry", io: IO) -> int:
        """To be implemented by workspace subcommands."""
        raise NotImplementedError()

    def post_handle(self) -> int:
        return 0

    def selected_projects(self, include_external: bool = False) -> List["Package"]:
        if self.option("since"):
            if self.option("project"):
                self.line("Both --project and --since flags are provided, using --since", style="warning")
            return self.changed_projects(include_external)

        return self.workspace.graph.search(
            package_names=self.option("project"),
            include_dependencies=self.option("include-dependencies"),
            include_reverse_dependencies=self.option("include-reverse-dependencies"),
            include_external=include_external,
        )

    def changed_projects(self, include_external: bool = False) -> List["Package"]:
        from poetry_workspace.diff import Diff

        ref = self.option("since")

        diff = Diff(self.workspace, io=self.io)
        changed = diff.get_changed_projects(ref)
        if self.option("include-reverse-dependencies"):
            changed.extend(diff.get_changed_external(ref))

        return self.workspace.graph.search(
            package_names=[package.name for package in changed],
            include_dependencies=self.option("include-dependencies"),
            include_reverse_dependencies=self.option("include-reverse-dependencies"),
            include_external=include_external,
        )

    def _io_for_project(self, name: str) -> IO:
        formatter = WorkspaceFormatter(name, decorated=self.io.output.is_decorated())

        # Shallow clone the outputs as we need to set different formatters for
        # each project.
        output = copy.copy(self.io.output)
        output.set_formatter(formatter)

        error_output = copy.copy(self.io.error_output)
        error_output.set_formatter(formatter)

        return IO(
            input=self.io.input,
            output=output,
            error_output=error_output,
        )



import os
import subprocess
from typing import TYPE_CHECKING, List, Tuple

from cleo.helpers import option
from poetry.console.commands.env_command import EnvCommand
from poetry.console.commands.run import RunCommand

from poetry_workspace.commands.workspace.workspace import WorkspaceCommand

if TYPE_CHECKING:
    from cleo.io.io import IO
    from poetry.poetry import Poetry


class WorkspaceRunCommand(WorkspaceCommand, EnvCommand):
    name = "workspace run"
    description = "Runs a command for each workspace project."

    options = [
        option("parallel", None, "Run all commands immediately."),
    ] + WorkspaceCommand.options
    arguments = RunCommand.arguments

    def __init__(self):
        super().__init__()

        # Used in parallel mode.
        self._procs: List[Tuple[subprocess.Popen, "IO"]] = []

    @property
    def parallel(self) -> str:
        return self.option("parallel")

    def handle_each(self, poetry: "Poetry", io: "IO") -> int:
        cmd = RunCommand()
        cmd.set_env(self.env)
        cmd.set_poetry(poetry)

        def execute(*args):
            command = self.env.get_command_from_bin(args[0]) + list(args[1:])
            env = dict(os.environ)

            proc = subprocess.Popen(
                args=command,
                cwd=poetry.file.path.parent,
                env=env,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
            if self.parallel:
                self._procs.append((proc, io))
            else:
                return self._communicate(proc, io)

        cmd.env.execute = execute
        return cmd.execute(io)

    def post_handle(self) -> int:
        if not self.parallel:
            return 0

        for proc, io in self._procs:
            exit_code = self._communicate(proc, io)
            if exit_code:
                return exit_code

        return 0

    def _communicate(self, proc: subprocess.Popen, io: "IO") -> int:
        stdout, stderr = proc.communicate()
        if stdout:
            io.write_line(stdout.strip())
        if stderr:
            io.write_line(stderr.strip())
        return proc.returncode



from typing import TYPE_CHECKING, List

from cleo.helpers import option

from poetry_workspace.commands.workspace.workspace import WorkspaceCommand

if TYPE_CHECKING:
    from cleo.io.io import IO
    from poetry.core.packages.package import Package
    from poetry.poetry import Poetry

_FORMATS = (
    "topological",
    "json",
    "tree",
)


class WorkspaceListCommand(WorkspaceCommand):
    name = "workspace list"
    description = "Lists workspace projects and their dependencies."
    help = """
The <info>workspace list</> command lists projects in the workspace's dependency
graph. Use flags to optionally select a subset of projects, and to output
their dependencies, reverse dependencies, and whether to include projects
external to the workspace in the dependency list."""

    options = [
        option("output", "o", f"Output format ({', '.join(_FORMATS)}).", flag=False, default=_FORMATS[0]),
        option("show-external", None, "Show external dependencies in the output."),
    ] + WorkspaceCommand.options

    def __init__(self):
        super().__init__()

        # Used for json and tree output formats.
        self._project_tree = {}

    @property
    def output(self) -> str:
        return self.option("output")

    @property
    def show_external(self) -> bool:
        return self.option("show-external")

    def handle(self) -> int:
        if not self.workspace:
            self.line("The 'workspace' command is only supported from within a workspace", style="error")
            return 1

        if self.output not in _FORMATS:
            self.line("unknown output format", style="error")
            return 1

        for project in self.selected_projects():
            if self.output == "topological":
                self.line(project.name)
                continue

            def get_tree(package_name: str) -> dict:
                deps = self.workspace.graph.dependencies(package_name)
                if not self.show_external:
                    deps = [dep for dep in deps if self.workspace.graph.is_project_package(dep)]
                return {dep.name: get_tree(dep.name) for dep in deps}

            self._project_tree[project.name] = get_tree(project.name)

        if self.output == "json":
            import json

            self.line(json.dumps(self._project_tree, indent=2))
        elif self.output == "tree":
            for project_name, tree in self._project_tree.items():
                self.write_tree(project_name, tree, [])
                self.line("")

        return 0

    def write_tree(self, project_name: str, tree: dict, last_levels: List[bool]) -> None:
        for i, is_last in enumerate(last_levels):
            if i != len(last_levels) - 1:
                if not is_last:
                    self.io.write("│   ")
                else:
                    self.io.write("    ")
            else:
                if not is_last:
                    self.io.write("├── ")
                else:
                    self.io.write("└── ")

        self.line(project_name)
        for i, (dep_name, dep_tree) in enumerate(tree.items()):
            self.write_tree(dep_name, dep_tree, last_levels + [i == len(tree) - 1])

    def selected_projects(self, *args) -> List["Package"]:
        projects = super().selected_projects(self.show_external)
        if self.output in ("json", "tree"):
            projects = sorted(projects, key=lambda project: project.name)
        return projects



import tempfile
from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional

from poetry.factory import Factory

from poetry_workspace.errors import VCSError
from poetry_workspace.graph import DependencyGraph
from poetry_workspace.vcs import VCS, detect_vcs
from poetry_workspace.workspace import Workspace

if TYPE_CHECKING:
    from cleo.io.io import IO
    from poetry.core.packages.package import Package


class Diff:
    _workspace: Workspace
    _vcs: VCS
    _io: "IO"

    def __init__(self, workspace: Workspace, vcs: VCS = None, io: "IO" = None):
        self._workspace = workspace
        self._vcs = vcs or detect_vcs()

        if io is None:
            from cleo.io.null_io import NullIO

            io = NullIO()
        self._io = io

    def get_changed_projects(self, ref: str) -> List["Package"]:
        package_paths = [(package, Path(package.source_url)) for package in self._workspace.graph.search()]

        def find_parent_project(file_path: Path) -> Optional["Package"]:
            for package, package_path in package_paths:
                try:
                    file_path.relative_to(package_path)
                    return package
                except ValueError:
                    pass
            return None

        changed_files = self.get_changed_files(ref)
        changed_projects = defaultdict(list)
        other_files = []

        for file in changed_files:
            package = find_parent_project(file)
            if package:
                changed_projects[package].append(file)
            else:
                other_files.append(file)

        if self._io.is_very_verbose():
            for package, files in changed_projects.items():
                self._io.write_line(f"Detected changes in project <info>{package.source_url}</info>:")
                for file in sorted(files):
                    self._io.write_line(f"- {file}")

            if other_files:
                self._io.write_line("Changed files not belonging to any project:")
                for file in sorted(other_files):
                    self._io.write_line(f"- {file}")

        return [package for package, _ in package_paths if package in changed_projects]

    def get_changed_external(self, ref: str) -> List["Package"]:
        old_graph = self.get_old_graph(ref)
        changed_external = []

        for package in self._workspace.graph:
            if self._workspace.graph.is_project_package(package):
                continue

            if not old_graph.has_package(package):
                changed_external.append(package)

        if self._io.is_very_verbose():
            self._io.write_line("Changed external packages:")
            for package in changed_external:
                self._io.write_line(f"- {package}")

        return changed_external

    def get_changed_files(self, ref: str) -> List[Path]:
        files = self._vcs.get_changed_files(ref)
        if self._io.is_very_verbose():
            self._io.write_line(f"Files changed since <info>{ref}</info>:")
            for file in files:
                self._io.write_line(f"- {file}")
        return files

    def get_old_graph(self, ref: str) -> DependencyGraph:
        old_pyproject_toml = self._vcs.read_file(ref, self._workspace.poetry.file.path)
        old_poetry_lock = self._vcs.read_file(ref, self._workspace.poetry.locker.lock.path)

        with tempfile.TemporaryDirectory() as temp_dir:
            dir_path = Path(temp_dir)
            (dir_path / "pyproject.toml").write_text(old_pyproject_toml)
            (dir_path / "poetry.lock").write_text(old_poetry_lock)

            poetry = Factory().create_poetry(dir_path)
            workspace_dir = self._workspace.poetry.file.parent

            # When creating the locked repository below, Poetry expects to find the
            # pyproject.toml or setup.py file for each of the directory dependencies.
            # Here we iterate through the lock file and try to populate those package
            # files from the old VCS reference.
            for package_info in poetry.locker.lock_data["package"]:
                source = package_info.get("source", {})
                if source.get("type") != "directory":
                    continue

                package_path = Path(source.get("url")) / "pyproject.toml"
                package_contents = None
                try:
                    package_contents = self._vcs.read_file(ref, workspace_dir / package_path)
                except VCSError:
                    pass

                if not package_contents:
                    package_path = Path(source.get("url")) / "setup.py"
                    try:
                        package_contents = self._vcs.read_file(ref, workspace_dir / package_path)
                    except VCSError:
                        pass

                if not package_contents:
                    raise VCSError(f"package {package_info['name']} at {ref} does not seem to be a Python package")

                temp_package_path = temp_dir / package_path
                temp_package_path.parent.mkdir(parents=True)
                temp_package_path.write_text(package_contents)

            locked_repo = poetry.locker.locked_repository(with_dev_reqs=True)
            return DependencyGraph(locked_repo, [])



from pathlib import Path

from poetry_workspace.errors import VCSError
from poetry_workspace.vcs.vcs import VCS


def detect_vcs() -> VCS:
    cwd = Path.cwd()
    for dir_path in [cwd] + list(cwd.parents):
        if (dir_path / ".git").is_dir():
            from poetry_workspace.vcs.git import Git

            return Git(dir_path)

    raise VCSError("unable to detect VCS type")



import json
from pathlib import Path

from poetry.core import json as poetry_json

SCHEMA_FILE_NAME = "poetry-schema.json"


def main():
    poetry_schema_path = Path(poetry_json.SCHEMA_DIR) / SCHEMA_FILE_NAME
    if not poetry_schema_path.exists():
        raise ValueError(f"{poetry_schema_path} does not exist")

    content = json.loads(poetry_schema_path.read_text())
    content["properties"]["workspace"] = {
        "type": "object",
        "description": "Information about the Poetry workspace.",
        "additionalProperties": False,
        "properties": {
            "include": {
                "type": "array",
                "description": "A list of folders to include.",
            },
            "exclude": {
                "type": "array",
                "description": "A list of folders to exclude.",
            },
        },
    }

    plugin_schema_path = Path(__file__).parent / SCHEMA_FILE_NAME
    plugin_schema_path.write_text(json.dumps(content, indent=2))


if __name__ == "__main__":
    main()



from poetry.console.commands.build import BuildCommand as BaseBuildCommand


class BuildCommand(BaseBuildCommand):
    def handle(self) -> int:
        from poetry.core.packages.dependency import Dependency
        from poetry.core.packages.directory_dependency import DirectoryDependency
        from poetry.core.pyproject.toml import PyProjectTOML

        # Replace directory dependencies specified by a path with one specified
        # by a version constraint.
        for i, dep in enumerate(self.poetry.package.requires):
            if not isinstance(dep, DirectoryDependency):
                continue

            pyproject_toml = PyProjectTOML(dep.full_path / "pyproject.toml")
            if pyproject_toml.is_poetry_project():
                version = pyproject_toml.poetry_config.get("version")
                if not version:
                    self.io.write_line(
                        f"<warning>No version property in project {dep.full_path}, using version '*'</warning>"
                    )
                    version = "*"
            else:
                self.io.write_line(f"<warning>Not a Poetry project {dep.full_path}, using version '*'</warning>")
                version = "*"

            self.poetry.package.requires[i] = Dependency(
                name=dep.name,
                constraint=version,
                optional=dep.is_optional(),
                groups=list(dep.groups),
                allows_prereleases=dep.allows_prereleases(),
                extras=dep.extras,
                source_type=dep.source_type,
                source_url=dep.source_url,
                source_reference=dep.source_reference,
                source_resolved_reference=dep.source_resolved_reference,
            )

        return super().handle() or 0



from typing import TYPE_CHECKING, List

from poetry.console.commands.publish import PublishCommand

from poetry_workspace.commands.workspace.workspace import WorkspaceCommand

if TYPE_CHECKING:
    from cleo.io.inputs.option import Option
    from cleo.io.io import IO
    from poetry.poetry import Poetry


def strip_flag_shortcuts(options: List["Option"]) -> List["Option"]:
    for opt in options:
        opt._shortcut = None
    return options


class WorkspacePublishCommand(WorkspaceCommand):
    name = "workspace publish"
    description = "Publishes workspace projects."

    options = strip_flag_shortcuts(PublishCommand.options) + WorkspaceCommand.options

    def handle_each(self, poetry: "Poetry", io: "IO") -> int:
        cmd = PublishCommand()
        cmd.set_poetry(poetry)
        return cmd.execute(io)



import subprocess
from pathlib import Path
from typing import List

from poetry_workspace.errors import VCSError
from poetry_workspace.vcs.vcs import VCS


class Git(VCS):
    def get_changed_files(self, ref: str) -> List[Path]:
        try:
            proc = subprocess.run(
                ["git", "diff", "--name-only", ref], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
        except subprocess.CalledProcessError as e:
            raise VCSError(f"git command failed:\n\n{e.stderr.decode()}")

        return [self.root / file for file in proc.stdout.decode().split()]

    def read_file(self, ref: str, file: Path) -> str:
        if file.is_absolute():
            file = file.relative_to(self.root)

        try:
            proc = subprocess.run(
                ["git", "show", f"{ref}:{file}"], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
        except subprocess.CalledProcessError as e:
            raise VCSError(f"git command failed:\n\n{e.stderr.decode()}")

        return proc.stdout.decode()



from typing import TYPE_CHECKING

from poetry.console.commands.build import BuildCommand
from poetry.console.commands.env_command import EnvCommand

from poetry_workspace.commands.workspace.workspace import WorkspaceCommand

if TYPE_CHECKING:
    from cleo.io.io import IO
    from poetry.poetry import Poetry


class WorkspaceBuildCommand(WorkspaceCommand, EnvCommand):
    name = "workspace build"
    description = "Builds workspace projects."

    options = BuildCommand.options + WorkspaceCommand.options
    loggers = BuildCommand.loggers

    def handle_each(self, poetry: "Poetry", io: "IO") -> int:
        cmd = BuildCommand()
        cmd.set_env(self.env)
        cmd.set_poetry(poetry)
        return cmd.execute(io)



from pathlib import Path
from typing import List


class VCS:
    def __init__(self, root: Path):
        self.root = root.resolve()

    def get_changed_files(self, ref: str) -> List[Path]:
        raise NotImplementedError()

    def read_file(self, ref: str, file: Path) -> str:
        raise NotImplementedError()



from typing import TYPE_CHECKING

from poetry.console.commands.version import VersionCommand

from poetry_workspace.commands.workspace.workspace import WorkspaceCommand

if TYPE_CHECKING:
    from cleo.io.io import IO
    from poetry.poetry import Poetry


class WorkspaceVersionCommand(WorkspaceCommand):
    name = "workspace version"
    description = "Shows or bumps the versions of workspace projects."

    arguments = VersionCommand.arguments
    options = VersionCommand.options + WorkspaceCommand.options

    def handle_each(self, poetry: "Poetry", io: "IO") -> int:
        cmd = VersionCommand()
        cmd.set_poetry(poetry)
        return cmd.execute(io)



from collections import defaultdict
from typing import TYPE_CHECKING, Dict, Iterator, List, Set

from poetry_workspace.errors import GraphError

if TYPE_CHECKING:
    from poetry.core.packages.package import Package
    from poetry.repositories import Repository


class DependencyGraph:
    _repo: "Repository"
    _internal_urls: Set[str]
    _deps: Dict["Package", List["Package"]]
    _rdeps: Dict["Package", List["Package"]]
    _levels: Dict["Package", int]
    _sorted: List["Package"]

    def __init__(self, repo: "Repository", internal_urls: List[str]):
        self._repo = repo
        self._internal_urls = set(internal_urls)
        self._deps = {}
        self._rdeps = {}

        # Ensure that all keys exist.
        for package in repo.packages:
            self._deps[package] = []
            self._rdeps[package] = []

        for package in repo.packages:
            for dep in package.all_requires:
                found = repo.find_packages(dep)
                if len(found) == 0:
                    raise ValueError(f"no packages found for dependency {dep.name}")
                if len(found) > 1:
                    raise ValueError(f"multiple packages found for dependency {dep.name}")

                self._deps[package].append(found[0])
                self._rdeps[found[0]].append(package)

        self._levels = topological_sort(self._deps, self._rdeps)

        sorted_levels = sorted(self._levels.items(), key=lambda pair: (pair[1], pair[0].name))
        self._sorted = [package for package, _level in sorted_levels]

    def __iter__(self) -> Iterator["Package"]:
        return iter(self._sorted)

    def __len__(self) -> int:
        return len(self._sorted)

    def has_package(self, package: "Package") -> bool:
        found = self._repo.find_packages(package.to_dependency())
        return len(found) > 0

    def is_project_package(self, package: "Package") -> bool:
        return package.source_url in self._internal_urls

    def dependencies(self, name: str) -> List["Package"]:
        return self._deps[self.find_package(name)]

    def reverse_dependencies(self, name: str) -> List["Package"]:
        return self._rdeps[self.find_package(name)]

    def search(
        self,
        package_names: List[str] = None,
        include_dependencies: bool = False,
        include_reverse_dependencies: bool = False,
        include_external: bool = False,
    ) -> List["Package"]:
        if package_names is None:
            if include_external:
                return self._sorted
            return [package for package in self._sorted if self.is_project_package(package)]

        selected = set()
        for name in package_names:
            selected.add(self.find_package(name))

        def add_dep(acc: Set["Package"], package: "Package", deps: Dict["Package", List["Package"]]) -> None:
            if package in acc:
                return
            acc.add(package)
            for dep in deps[package]:
                add_dep(acc, dep, deps)

        dependencies: Set["Package"] = set()
        if include_dependencies:
            for package in selected:
                add_dep(dependencies, package, self._deps)

        reverse_dependencies: Set["Package"] = set()
        if include_reverse_dependencies:
            for package in selected:
                add_dep(reverse_dependencies, package, self._rdeps)

        results = selected.union(dependencies).union(reverse_dependencies)
        if not include_external:
            results = {package for package in results if self.is_project_package(package)}

        return sorted(results, key=lambda package: (self._levels[package], package.name))

    def find_package(self, name: str) -> "Package":
        results = self._repo.search(name)
        if not results:
            raise GraphError(f"Project '{name}' is not in the dependency graph")
        return results[0]


def topological_sort(
    deps: Dict["Package", List["Package"]],
    rdeps: Dict["Package", List["Package"]],
) -> Dict["Package", int]:
    """
    Returns a dictionary mapping a package to its level. Levels are strictly
    negative integers that represent the maximum number of connections from
    the root to the package, e.g. -1 is the highest possible level and means
    that it is a direct dependency of the root and no other transitive package
    depends upon it.
    """
    levels: Dict["Package", int] = defaultdict(int)

    def iter(package: "Package", level: int) -> None:
        if levels[package] <= level:
            return
        levels[package] = level
        for dep in deps[package]:
            iter(dep, level - 1)

    for package, package_rdeps in rdeps.items():
        if len(package_rdeps) == 0:
            # Begin iteration at top level deps.
            iter(package, -1)

    return levels



from importlib import import_module
from typing import TYPE_CHECKING, Callable

if TYPE_CHECKING:
    from poetry.console.command_loader import CommandLoader
    from poetry.console.commands.command import Command

COMMAND_NAMES = [
    "build",
    "workspace build",
    "workspace list",
    "workspace publish",
    "workspace run",
    "workspace version",
]


def register_commands(command_loader: "CommandLoader") -> None:
    def load_plugin_command(name: str) -> Callable[[], "Command"]:
        def _load() -> "Command":
            module = import_module("poetry_workspace.commands.{}".format(".".join(name.split(" "))))
            command_class = getattr(module, "{}Command".format("".join(c.title() for c in name.split(" "))))
            return command_class()

        return _load

    for name in COMMAND_NAMES:
        # Set _factories directly instead of calling the register_factory method
        # because we want to override Poetry's base commands.
        command_loader._factories[name] = load_plugin_command(name)



from cleo.formatters.formatter import Formatter


class WorkspaceFormatter(Formatter):
    def __init__(self, prefix: str, *args, **kwargs):
        self._prefix = prefix
        super().__init__(*args, **kwargs)

    def format(self, message: str) -> str:
        lines = message.split("\n")
        return super().format("\n".join(f"<c2>{self._prefix}></> {line}" for line in lines))



import os
from glob import glob
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Set

from poetry.core import json
from poetry.core.packages.directory_dependency import DirectoryDependency
from poetry.factory import Factory

from poetry_workspace.errors import WorkspaceError
from poetry_workspace.graph import DependencyGraph

if TYPE_CHECKING:
    from cleo.io.io import IO
    from poetry.core.pyproject.toml import PyProjectTOML
    from poetry.poetry import Poetry


class Workspace:
    _pyproject: "PyProjectTOML"
    _poetry: "Poetry"
    _projects: List["Poetry"]
    _io: "IO"
    _graph: Optional[DependencyGraph]

    def __init__(self, pyproject: "PyProjectTOML", io: "IO"):
        if not is_workspace_pyproject(pyproject):
            raise WorkspaceError("pyproject.toml file does not contain a 'tool.poetry.workspace' section")

        monkeypatch_json_schema()

        self._pyproject = pyproject
        self._poetry = Factory().create_poetry(pyproject.file.path)
        self._io = io
        self._graph = None

        self._projects = self._find_projects(pyproject)
        self._add_project_dependencies()

    @property
    def poetry(self) -> "Poetry":
        return self._poetry

    @property
    def projects(self) -> List["Poetry"]:
        return self._projects

    @property
    def graph(self) -> DependencyGraph:
        if self._graph is None:
            locked_repo = self.poetry.locker.locked_repository(with_dev_reqs=True)
            internal_urls = [str(project.file.path.parent) for project in self.projects]
            self._graph = DependencyGraph(locked_repo, internal_urls)
        return self._graph

    def find_project(self, name: str) -> Optional["Poetry"]:
        for project in self.projects:
            if project.package.name == name:
                return project
        return None

    def _find_projects(self, pyproject: "PyProjectTOML") -> List["Poetry"]:
        content = pyproject.data["tool"]["poetry"]["workspace"]
        if "include" not in content:
            raise WorkspaceError("pyproject.toml file requires 'include' in the 'tool.poetry.workspace' section")

        include = content["include"]
        exclude = content.get("exclude", [])

        matches: Set[str] = set()
        for pattern in include:
            pattern = str(self.poetry.file.path.parent / pattern / "pyproject.toml")
            matches = matches.union(set(glob(pattern, recursive=True)))

        for pattern in exclude:
            pattern = str(self.poetry.file.path.parent / pattern / "pyproject.toml")
            matches = matches.difference(set(glob(pattern, recursive=True)))

        if self._io.is_very_verbose():
            self._io.write_line(f"Using workspace {self.poetry.file.path}")
            self._io.write_line("Found workspace projects:")
            for path in sorted(matches):
                self._io.write_line(f"- {path}")

        return [Factory().create_poetry(Path(path)) for path in sorted(matches)]

    def _add_project_dependencies(self) -> None:
        requires = set(pkg.name for pkg in self.poetry.package.requires)

        for project in self.projects:
            name = project.pyproject.poetry_config["name"]
            if name in requires:
                continue

            self.poetry.package.add_dependency(
                DirectoryDependency(
                    name=name,
                    path=project.file.path.parent,
                    develop=True,
                )
            )


def is_workspace_pyproject(pyproject: "PyProjectTOML") -> bool:
    return pyproject.file.exists() and pyproject.data.get("tool", {}).get("poetry", {}).get("workspace") is not None


def monkeypatch_json_schema() -> None:
    """
    Monkeypatch Poetry's JSON schema for the pyproject.toml file with our custom
    one that includes the schema for `tool.poetry.workspace` section. See
    schemas/gen_schema.py for details.
    """
    json.SCHEMA_DIR = os.path.join(os.path.dirname(__file__), "schemas")
